package com.blackledge.david.html.hme;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.net.CookieHandler;
import java.net.CookieManager;
import java.net.CookieStore;
import java.net.HttpCookie;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map.Entry;
import java.util.regex.Pattern;

import com.blackledge.david.html.ThreadLocalCookieStore;
import com.blackledge.david.html.hme.host.HMEItem;
import com.blackledge.david.html.hme.host.HostContext;

import com.tivo.core.ds.TeDict;
import com.tivo.hme.interfaces.IContext;

/**
 * Handle Bookmark storage and Cookies for WebBrowserApp.<BR/>
 * Uses System.getProperty("links", "default_links.txt"). Its entries tab
 * separated are: <BR/>
 * URL Folder\rFolder\rTitle Icon yyyy-MM-dd@HH:mm:ss<BR/>
 * New (past) Dated entries will be added to existing lists, replacing a first same-URL
 * entry, deleting it if the Title is blank.
 */
public class Bookmarks {
	
	static final String BOOKMARK_FOLDER_PREFIX = "file:"; //$NON-NLS-1$
	static final String BOOKMARK_FOLDER_SEPARATOR = "\r"; //$NON-NLS-1$
	
	private static final String BOOKMARK_ENTRY_SEP = "\n"; //$NON-NLS-1$
	private static final String BOOKMARK_URL_TITLE_SEP = "\t"; //$NON-NLS-1$
	private static final String BOOKMARKS_KEY = "BOOKMARKS"; //$NON-NLS-1$

	private WebBrowserApp _app;

	public Bookmarks(WebBrowserApp app) {
		_app = app;
	}
	
	public IContext getContext() {
		return _app.getContext();
	}

	private void logBookmark(String what, String url, String icon,
			String folder, String titleString) {
		_app.logBookmark(what, url, icon, folder, titleString);
	}

	private boolean isBookmarksDisabled() {
		return _app.isBookmarksDisabled();
	}

	public boolean saveBookmark(URL urlBase, String titleString, String folder, URL icon) {
		// apply folder to titlestring
		String oldFolder = getFolder(titleString);
		// remove old folder
		if (oldFolder != null) {
			titleString = titleString.substring(oldFolder.length());
		}
		if (folder.startsWith(BOOKMARK_FOLDER_PREFIX)) {
			folder = folder.substring(BOOKMARK_FOLDER_PREFIX.length());
		}
		// add new folder
		if (!folder.endsWith(BOOKMARK_FOLDER_SEPARATOR) && folder.length() > 0) {
			folder = folder + BOOKMARK_FOLDER_SEPARATOR;
		}
		
		logBookmark("thumztakedit", urlBase==null?null:urlBase.toString(), icon==null?null:icon.toString(), folder, titleString);
		
		titleString = folder + titleString;
		// save bookmark
		return saveBookmark(urlBase, titleString, icon);
	}

	public boolean saveBookmark(URL urlBase, String titleString, URL icon) {
		if (!isBookmarksDisabled()) {
			String bookmarkString = getContext().getPersistentData(
					BOOKMARKS_KEY);
			saveBookmark(urlBase, titleString, icon, bookmarkString);
			return true;
		}
		return false;
	}
	
	public String saveBookmark(URL urlBase, String titleString, URL icon, String bookmarkString) {
		if (!isBookmarksDisabled()) {
			if (bookmarkString == null) {
				bookmarkString = ""; //$NON-NLS-1$
			}
			if(bookmarkString.trim().length() == 0) {
				bookmarkString = addDefaultBookmarks(bookmarkString);
			}
			bookmarkString = addBookmark(bookmarkString, urlBase, titleString,
					icon);
			getContext().setPersistentData(BOOKMARKS_KEY, bookmarkString);
			return bookmarkString;
		} else {
			return bookmarkString;
		}
	}

	private String addDefaultBookmarks(String bookmarkString) {
		LinkedHashMap<String, URL[]> bookmarks = new LinkedHashMap<String, URL[]>();
		// oh... ok, add all the default bookmarks so they don't lose them.
		try {
			addDefaultWebsiteBookmarks(bookmarks);
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		for(String title : bookmarks.keySet()) {
			URL urls[] = bookmarks.get(title);
			if(urls == null || urls.length < 2)
				continue;
			URL url = urls[0];
			URL iconUrl = urls[1];
			bookmarkString = addBookmark(bookmarkString, url, title, iconUrl);
		}
		return bookmarkString;
	}

	public boolean removeBookmark(URL urlBase, String titleString, URL icon) {
		if (!isBookmarksDisabled()) {
			String bookmarkString = getContext().getPersistentData(
					BOOKMARKS_KEY);
			if (bookmarkString == null) {
				bookmarkString = ""; //$NON-NLS-1$
			}
			bookmarkString = removeBookmark(bookmarkString, urlBase,
					titleString);
			getContext().setPersistentData(BOOKMARKS_KEY, bookmarkString);
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Use the returned string in any methods that require the bookmark string
	 * data as an argument when you want to use a cached copy to improve
	 * performance.
	 * 
	 * @param context
	 * @return
	 */
	public static String getBookmarkData(IContext context) {
		return context.getPersistentData(BOOKMARKS_KEY);
	}
	public LinkedHashMap<String, URL[]> getBookmarks() {
		return getBookmarks(null);
	}
	public LinkedHashMap<String, URL[]> getBookmarks(String folder) {
		String bookmarkString = getBookmarkData(getContext());
		return getBookmarks(folder, bookmarkString, null);
	}
	
	/**
	 * Resulting titles have folder removed from the front. This is called with
	 * folder WebBrowserApp.readablePathToFolder("Apps/") as well as
	 * WebBrowserApp.readablePathToFolder("Apps/genre/") to get the TiVoConnect
	 * list.
	 * 
	 * @param folder
	 * @param bookmarkString
	 * @return
	 */
	public static LinkedHashMap<String, URL[]> getBookmarks(String folder, String bookmarkString, IContext ctx) {
		if(folder != null && folder.startsWith(BOOKMARK_FOLDER_PREFIX)) {
			folder = getFolder(folder);
		}
		LinkedHashMap<String, URL[]> result = new LinkedHashMap<String, URL[]>();
		if (bookmarkString != null) {
			getBookmarks(result, bookmarkString);
		}

		if(result.size() == 0) {
			LinkedHashMap<String, URL[]> bookmarks = getDefaultBookmarks();

			result.putAll(bookmarks);
		} else if(ctx != null && ctx instanceof HostContext) {
			applyBookmarkUpdates(result, getBookmarksDate((HostContext) ctx));
		}
		
		ArrayList<String> subfolders = new ArrayList<String>();
		if(folder != null) {
			Iterator<Entry<String, URL[]>> entries = result.entrySet().iterator();
			while(entries.hasNext()) {
				Entry<String, URL[]> bookmark = entries.next();
				// strip entries that don't start with folder
				if(!bookmark.getKey().startsWith(folder)) {
					entries.remove();
					continue;
				}
				String subfolder = getFolder(bookmark.getKey().substring(folder.length()));
				// strip entries that are in a subfolder (just leave a folder entry behind)
				if(subfolder != null) {
					entries.remove();
					if(!subfolders.contains(subfolder))
						subfolders.add(subfolder);
					continue;
				}
			}
			LinkedHashMap<String, URL[]> nonFolders = result;
			result = new LinkedHashMap<String, URL[]>();
			if(subfolders.size() > 0) {
				// put subfolders at beginning
				for(String subfolder : subfolders) {
					//FIXME define urls for this case
					try {
						result.put(subfolder, new URL[] {new URL(BOOKMARK_FOLDER_PREFIX+folder+subfolder),null});
					} catch (MalformedURLException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			}
			for(Entry<String, URL[]>bookmark : nonFolders.entrySet()) {
				result.put(bookmark.getKey().substring(folder.length()), bookmark.getValue());
			}
		}
		return result;
	}
	private static LinkedHashMap<String, URL[]> getDefaultBookmarks() {
		LinkedHashMap<String, URL[]> bookmarks = new LinkedHashMap<String, URL[]>();
		try {
			addDefaultWebsiteBookmarks(bookmarks);
			
//			addDemoBookmarks(bookmarks);
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return bookmarks;
	}

	public static void addDemoBookmarks(LinkedHashMap<String, URL[]> bookmarks)
			throws MalformedURLException {
		URL url;
		URL icon;
		url = new URL("hme://wmcbrine.com/reversi"); //$NON-NLS-1$
		icon = new URL("http://wmcbrine.com/images/reversi.png"); //$NON-NLS-1$
		bookmarks.put("Games"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch Reversi game by wmcbrine", new URL[] {url, icon}); //$NON-NLS-1$
			
//		url = new URL("hme://168.143.122.49:7288/mp3tunes");
		url = new URL("hme://www.mp3tunes.com"); // points to .49 ip //$NON-NLS-1$
		icon = new URL("http://168.143.122.40:7288/mp3tunes/icon.png"); //$NON-NLS-1$
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+HMEItem.GENRE_MUSIC+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch ('hme://...') MP3Tunes App [:( unplugged 12/2012]", new URL[] {url, icon}); //$NON-NLS-1$

//		url = new URL("http://sideload.com/#primarynav");
//		icon = new URL("http://s.mp3tunes.com/images/locker/sideload_logo.gif");
//		bookmarks.put("(temp)sideload.com - add free MP3s to MP3Tunes", new URL[] {url, icon});

		// ticketchannel.tv still serves TiVoConnect, but no app at pointed location
//		url = new URL("hme://69.72.234.59:7288/hme");
//		url = new URL("hme://69.72.234.62:7288/hme"); // listed in ticketchannel simulator jnlp launcher
//		url = new URL("hme://ticketchannel.tv"); //points to .59 ip
//		bookmarks.put("(temp)TicketChannel TiVo App", new URL[] {url, icon});

//		// apps.tv configures based on TiVo id, we're not a tivo.
//		// also, the website is auctioned off as of 9/12/12
//		url = new URL("hme://209.97.196.52:7288/submanager");
////			url = new URL("hme://ns.apps.tv"); // or mail.apps.tv or mkloop.com (convert graphic to 1-frame mpg!)
//		// only mkloop.com exists as of 9/3/2012, and that won't give you TiVoConnect
//		//however hme handler now converts host to ip address before it begins.
//		// mkloop.com gone as of 9/6/2012
//		// no names available for apps.tv, and app is gone, now, too
//		bookmarks.put("(temp)launch apps.tv App", new URL[] {url, icon});

//		url = new URL("hme://204.176.49.116:7288/CreditCard");
//		icon = null;
//		bookmarks.put("(temp)launch TiVo's (hidden) Credit Card Update (PCI enabled) App", new URL[] {url, icon});
//		url = new URL("hme://204.176.49.116:7288/TivoUpgrade");
//		icon = null;
//		bookmarks.put("(temp)launch TiVo's Buy a TiVo Premiere App", new URL[] {url, icon});
		url = new URL("hme://204.176.49.116:7288/amazon"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"TiVo's Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's (hidden) Amazon App", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("hme://204.176.49.116:7288/bundle-netflix"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"TiVo's Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's (hidden) netflix App", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("hme://204.176.49.116:7288/dominos"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"TiVo's Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's (hidden) Order Domino's Pizza Now App", new URL[] {url, icon}); //$NON-NLS-1$
//		url = new URL("hme://204.176.49.116:7288/fandivo");
//		icon = null;
//		bookmarks.put("launch TiVo's Browse and Buy Movie Tickets App", new URL[] {url, icon});
//		url = new URL("hme://204.176.49.116:7288/hdphotos");
//		icon = null;
//		bookmarks.put("launch TiVo's Photos & Slideshows App", new URL[] {url, icon});
		url = new URL("hme://204.176.49.116:7288/hdui"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"TiVo's Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's (hidden) TiVo Search App (hdui)", new URL[] {url, icon}); //$NON-NLS-1$
//		url = new URL("hme://204.176.49.116:7288/hmedown");
//		icon = null;
//		bookmarks.put("launch TiVo's (hidden) Hme Down App", new URL[] {url, icon});
//		url = new URL("hme://204.176.49.116:7288/kidzone");
//		icon = null;
		bookmarks.put("launch TiVo's (hidden) Guru Guides App", new URL[] {url, icon});
		// same basic url for both
//		bookmarks.put("launch TiVo's (hidden) KidZone App", new URL[] {url, icon});
		url = new URL("hme://204.176.49.116:7288/launcher"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"TiVo's Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's TiVo Search Beta App", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("hme://204.176.49.116:7288/live365"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"TiVo's Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's Live365 App", new URL[] {url, icon}); //$NON-NLS-1$
			
//		url = new URL("hme://204.176.49.116:7288/musicchoice");
//		icon = null;
//		bookmarks.put("launch TiVo's Music Videos from Music Choice App", new URL[] {url, icon});
//		url = new URL("hme://204.176.49.116:7288/norhapsody");
//		icon = null;
//		bookmarks.put("launch TiVo's Rhapsody Music Service App placeholder (norhapsody)", new URL[] {url, icon});
			
		url = new URL("hme://204.176.49.116:7288/pandora"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"TiVo's Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's Pandora App", new URL[] {url, icon}); //$NON-NLS-1$
//		url = new URL("hme://204.176.49.116:7288/podcast");
//		icon = null;
//		bookmarks.put("launch TiVo's Podcaster App", new URL[] {url, icon});
		url = new URL("hme://204.176.49.116:7288/samegame"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Games"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's SameGame App", new URL[] {url, icon}); //$NON-NLS-1$
		url = new URL("hme://204.176.49.116:7288/skullbones"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Games"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's Skull & Bones App", new URL[] {url, icon}); //$NON-NLS-1$
//		url = new URL("hme://204.176.49.116:7288/survey");
//		icon = null;
//		bookmarks.put("launch TiVo's (hidden) Survey App", new URL[] {url, icon});
		url = new URL("hme://204.176.49.116:7288/swivel"); //$NON-NLS-1$
		icon = null;
		// two versions of this
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"TiVo's Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's (hidden) Swivel App", new URL[] {url, icon}); //$NON-NLS-1$
		url = new URL("hme://204.176.49.116:7288/tivocast"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"TiVo's Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's (hidden) TiVoCast App (video on demand)", new URL[] {url, icon}); //$NON-NLS-1$
//		url = new URL("hme://204.176.49.116:7288/videoplayer");
//		icon = null;
//		bookmarks.put("launch TiVo's (hidden) Video Player App", new URL[] {url, icon});
		url = new URL("hme://204.176.49.116:7288/wordsmith"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Games"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's WordSmith App", new URL[] {url, icon}); //$NON-NLS-1$
		url = new URL("hme://204.176.49.116:7288/youtube"); //$NON-NLS-1$
		icon = null;
		// two versions, second one labeled "tcast jump"
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"TiVo's Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"launch TiVo's (hidden) YouTube App", new URL[] {url, icon}); //$NON-NLS-1$
					
//		url = new URL("hme://204.176.49.116/flash/youtube_leanback");
//		icon = null;
//		bookmarks.put("(temp)launch FLASH version of TiVo's (hidden) YouTube App", new URL[] {url, icon});
					
		/*
		urls: <url>
		 https://204.176.49.59:7287/CreditCard_20080912-20080912_131340/1221250421/1289849840-1605209840/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.59:7287/TivoUpgrade-20120424_122953/1335295796/1335399021-1650759021/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/amazon-1-9.2011.03.16-0632-20110316_154048/1300315250/1317939394-1633299394/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/dominos.2009.05.07.r1021.logerror.signed-20090507_174812/1241743696/1289849802-1605209802/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/fandivo-1.1.2011.11.10-1645-20111115_112508/1321385109/1343238702-1658598702/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/hdphotos-ml_2010.09.13-1339-20100913_134517/1284410719/1343239081-1658599081/clientip_67.0.59.230/
		</url>
		<url>
		 http://204.176.49.116:7288/hmedown_20080317-20080402_170823/1207181304/1240249987-1587405187/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/kidzone_v1.0-20100622-20100622_141902/1277241543/1289849391-1605209391/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/kidzone_v1.0-20100622-20100622_141902/1277241543/1289849349-1605209349/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/live365-1.0_2010.08.02-2340-20100802_234348/1280817829/1343239177-1658599177/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/musicchoice-1.0-2010.08.04-0632-20100804_161353/1280963634/1328046275-1643406275/clientip_67.0.59.230/
		</url>
		<url>
		 http://204.176.49.116:7288/podcast-20100730-qbert-20100730_180250/1280538174/1343239384-1658599384/clientip_67.0.59.230/
		</url>
		<url>
		 http://204.176.49.116:7288/samegame_20071012_01-20080402_173700/1207183021/1343239500-1658599500/clientip_67.0.59.230/
		</url>
		<url>
		 http://204.176.49.116:7288/skullbones_20071012-20080402_173743/1207183066/1343239554-1658599554/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/survey_20080530-01_signed-20080530_150815/1212185297/1289848835-1605208835/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/swivel-2.8.3-2012-05-08-0603at553106-20120508_094009/1336495211/1336595307-1651955307/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/swivel-2.8.3-2012-05-08-0603at553106-20120508_094009/1336495211/1336595232-1651955232/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/tivocast-2.8-2012-06-26-0603at563743-20120626_120132/1340737293/1344881428-1660241428/clientip_67.0.59.230/
		</url>
		<url>
		 https://204.176.49.65:7287/videoplayer-0023-release-sst-20090716_122445/1247772287/1333658049-1649018049/clientip_67.0.59.230/
		</url>
		<url>
		 http://204.176.49.116:7288/wordsmith_20071012-20080402_180020/1207184421/1343239627-1658599627/clientip_67.0.59.230/
		</url>			
		 */
	}

	public static void addDefaultWebsiteBookmarks(LinkedHashMap<String, URL[]> bookmarks)
			throws MalformedURLException {
		if(addDefaultWebsiteBookmarksFromFile(bookmarks)) {
			return;
		}
		
		URL url;
		URL icon;
		url = new URL("https://docs.google.com/document/pub?id=1o0jVDy2Ek-baZ7sm3v9OtTxZ9MksS-7RzLCY_1GU51w"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("[About "+WebBrowserApp.BROWSER_NAME+"]", new URL[]{url, icon}); //$NON-NLS-1$

		if(WebBrowserApp.getDebugLevel() > 0) {
			url = new URL("http://192.168.0.20:80/EnterWebzTest");
			icon = null;
			bookmarks.put("(temp)Test HTML Page", new URL[] {url, icon});
		}
		/*
 "type":"uiNavigate",
 "uiDestinationType":"hme",
 "uri":"x-tivo:hme:uuid:863cb78f-efdd-4106-b572-51733983dc76",
 "parameters":{
  "title":"Title goes here",
  "subtitle":"Subtitle goes here",
  "description":"Description goes here",
  "uri":"http://192.168.10.198:9032/videos/movies/star_wars2.mpg?Format=video%2Fx-tivo-mpeg",
  "duration":8280
 }
		 * */

//			url = new URL("hme://192.168.0.20:7288/webz/?url=http://voice.google.com&test=blah");
//			icon = null;
//			bookmarks.put("(temp)Test HME Transition", new URL[] {url, icon});
		
		// TODO look up host name to get IP, THEN make request for IP/TiVoConnect... that works in the old apps.tv case

		url = new URL("http://hmedev.wikidot.com/enter-webz-links#main-content"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("["+WebBrowserApp.BROWSER_NAME+" suggested links page]", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("http://www.google.com"); //$NON-NLS-1$
		icon = new URL("http://www.google.com/images/google_favicon_128.png"); //$NON-NLS-1$
		bookmarks.put("Google.com", //$NON-NLS-1$
						new URL[] {url, icon});
		
		url = new URL("http://www.youtube.com"); //$NON-NLS-1$
		icon = new URL("http://s.ytimg.com/yts/img/favicon_32-vflWoMFGx.png"); //$NON-NLS-1$
		bookmarks.put("YouTube.com", new URL[] {url, icon}); //$NON-NLS-1$
		
//			
//url = new URL("http://feeds.feedburner.com/ICanHasCheezburger");
//icon = null;
//bookmarks.put("rss cheezburger",  new URL[]{url, icon});
//
//url = new URL("http://feeds.feedburner.com/AnimalGifs");
//icon = null;
//bookmarks.put("rss animal gifs",  new URL[]{url, icon});

		url = new URL("http://en.m.wikipedia.org"); //$NON-NLS-1$
		icon = new URL("http://en.wikipedia.org/apple-touch-icon.png"); //$NON-NLS-1$
		bookmarks.put("Wikipedia.org (mobile)", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("http://m.wund.com/"); //$NON-NLS-1$
		icon = new URL("http://icons.wunderground.com/graphics/smalllogo.gif"); //$NON-NLS-1$
		bookmarks.put("(add zip, re-save)Weather Underground (mobile)", new URL[] {url, icon}); //$NON-NLS-1$
		
//			url = new URL("https://www.google.com/voice/m/i/voicemail");
		url = new URL("https://www.google.com/voice/m"); //$NON-NLS-1$
		icon = null;//new URL();
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"Google Voice", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("http://mail.google.com/mail/x/"); //$NON-NLS-1$
		icon = new URL("http://mail.google.com/mail/images/2/mail_icon_32.png"); //$NON-NLS-1$
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"Gmail (mobile)", new URL[] {url, icon}); //$NON-NLS-1$
		
		url=new URL("http://mail.google.com/mail/x/?ui=html"); //$NON-NLS-1$
		icon = new URL("http://mail.google.com/mail/images/2/mail_icon_32.png"); //$NON-NLS-1$
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"Gmail (basic HTML)", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("http://m.facebook.com/"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"Facebook.com (mobile)", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("http://imdb.com"); //$NON-NLS-1$
		icon = new URL("http://i.media-imdb.com/images/SF0ae8798307d147d8ea59044184b4339e/mobile/icon_114x114.png"); //$NON-NLS-1$
		bookmarks.put("Internet Movie Database (IMDB.com)", new URL[] {url, icon});			 //$NON-NLS-1$

		url = new URL("http://archive.org/details/feature_films"); //$NON-NLS-1$
		icon = new URL("http://archive.org/images/glogo.jpg"); //$NON-NLS-1$
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+HMEItem.GENRE_VIDEO+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"Feature Films at Archive.org", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("http://archive.org/details/etree"); //$NON-NLS-1$
		icon = new URL("http://archive.org/images/glogo.jpg"); //$NON-NLS-1$
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+HMEItem.GENRE_MUSIC+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"Live Concerts at Archive.org", new URL[] {url, icon}); //$NON-NLS-1$

		url = new URL("http://archive.org/details/classic_tv"); //$NON-NLS-1$
		icon = new URL("http://archive.org/images/glogo.jpg"); //$NON-NLS-1$
		bookmarks.put("Apps"+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+HMEItem.GENRE_VIDEO+Bookmarks.BOOKMARK_FOLDER_SEPARATOR+"Classic TV at Archive.org", new URL[] {url, icon}); //$NON-NLS-1$

		url = new URL("http://archive.org/details/americana"); //$NON-NLS-1$
		icon = new URL("http://archive.org/images/glogo.jpg"); //$NON-NLS-1$
		bookmarks.put("American Library texts at Archive.org", new URL[] {url, icon}); //$NON-NLS-1$

		url = new URL("http://tivocommunity.com"); //$NON-NLS-1$
		icon = new URL("http://www.tivocommunity.com/home/sites/default/files/tcflogonew5.png"); //$NON-NLS-1$
		bookmarks.put("TiVoCommunity.com", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("https://mobile.twitter.com/TiVo"); //$NON-NLS-1$
		icon = new URL("https://ma.twimg.com/twitter-mobile/e85c3330835e527a4e9f0f9621fc53841a65c1b8/images/ic_favicon.png"); //$NON-NLS-1$
		bookmarks.put("Twitter.com: TiVo (mobile)", new URL[] {url, icon}); //$NON-NLS-1$
		
		url = new URL("http://xkcd.com"); //$NON-NLS-1$
		icon = new URL("http://imgs.xkcd.com/static/apple-touch-icon-precomposed.png"); //$NON-NLS-1$
		bookmarks.put("xkcd.com comics", new URL[] {url, icon}); //$NON-NLS-1$
		
//			url = new URL("http://icanhascheezburger.com/#MobileSearch");
//			icon = null;
//			bookmarks.put("BROKEN LOLcats - i can has cheezburger?", new URL[] {url, icon});
		
		url = new URL("http://david.blackledge.com"); //$NON-NLS-1$
		icon = null;
		bookmarks.put("David.Blackledge.com", new URL[] {url, icon}); //$NON-NLS-1$
		
		addDemoBookmarks(bookmarks);
	}
	
	private static boolean applyBookmarkUpdates(java.util.LinkedHashMap<String, URL[]> bookmarks, Date bookmarkDate) {
		return addDefaultWebsiteBookmarksFromFile(bookmarks, bookmarkDate);		
	}
	
	private static Date getBookmarksDate(HostContext ctx) {
		return ctx.getPersistentFileLastModified();
	}
	
	private Date lastBookmarkDate = null;
	/**
	 * this is just to make sure we always use the same date, but it's really
	 * only called once ever during the session anyhow
	 */
	private Date getLastBookmarkDate(IContext ctx) {
		if(lastBookmarkDate == null) {
			if(ctx instanceof HostContext) {
				lastBookmarkDate = getBookmarksDate((HostContext) ctx);
			}
		}
		return lastBookmarkDate;
	}
	
	private static boolean addDefaultWebsiteBookmarksFromFile(java.util.LinkedHashMap<String, URL[]> bookmarks) {
		return addDefaultWebsiteBookmarksFromFile(bookmarks, null);
	}
	
	private static DateFormat linkDateFormat = null;
	
	private static boolean addDefaultWebsiteBookmarksFromFile(java.util.LinkedHashMap<String, URL[]> bookmarks, Date lastUpdateDate) {
		String filename = System.getProperty("links", "default_links.txt");
		BufferedReader reader;
		try {
			boolean changed = false;
			reader = new BufferedReader(new FileReader(filename));
			for (String line = reader.readLine(); line != null; line = reader.readLine()) {
				if (line.startsWith("#") || line.trim().length() == 0)continue; //$NON-NLS-1$
				String parts[] = line.split("\t"); //$NON-NLS-1$
				if (parts.length < 2)
					continue;
				String url_s = parts[0];
				String title = parts[1];
				title = title.replaceAll("\\\\r", BOOKMARK_FOLDER_SEPARATOR);
				String icon_s = null;
				if (parts.length > 2)
					icon_s = parts[2];

				boolean updating = false;
				if(lastUpdateDate != null) {
					String updateDateString = null;
					Date updateDate = null;
					if (parts.length > 3)
						updateDateString = parts[3];
					if(updateDateString != null && updateDateString.trim().length() > 0) {
						try {
							if(linkDateFormat == null) {
								linkDateFormat = new SimpleDateFormat("yyyy-MM-dd@HH:mm:ss");
								linkDateFormat.setLenient(true);
							}
							updateDate = linkDateFormat.parse(updateDateString);
						} catch (Exception e) {
							System.out.println("*** bad default_links.txt date \""+updateDateString+"\": "+e);
							//FIXME WebBrowserApp.log("id", "where", "configError", new String[] {"bad default_links.txt date \""+updateDateString+"\": "+e});
						}
					}
					// only add lines that have a past date and are newer than last update.
					if(updateDate == null || updateDate.before(lastUpdateDate) || updateDate.after(new Date())) {
						continue;
					}
					updating = true;
				}
				
				try {
					URL url = new URL(url_s);
					URL icon = null;
					if (icon_s != null && icon_s.trim().length() > 0) {
						try {
							icon = new URL(url, icon_s);
						} catch (Exception e) {
							WebBrowserApp.debugOut("Link icon Failed (" + icon_s + "): " + e);
							continue;
						}
					}

					if (updating) {
						// remove same-url entry first
						String deleteKey = null;
						for (String key : bookmarks.keySet()) {
							URL urls[] = bookmarks.get(key);
							if (urls.length > 0 && urls[0].equals(url)) {
								deleteKey = key;
								break;
							}
						}
						if (deleteKey != null) {
							bookmarks.remove(deleteKey);
							changed = true;
						}
						
						// just delete it if there is no title
						if(title.trim().length() == 0) {
							WebBrowserApp.debugOut("Deleted Link: " + line); //$NON-NLS-1$
							continue;
						}
					}
					
					bookmarks.put(title, new URL[] { url, icon });
					changed = true;
					WebBrowserApp.debugOut("Load Link: " + line); //$NON-NLS-1$
				} catch (Exception e) {
					WebBrowserApp.debugOut("Link Failed (" + line + "): " + e);
					continue;
				}
			}
			reader.close();
			return changed;
		} catch (FileNotFoundException e) {
			WebBrowserApp.debugOut("load Links "+filename+": " + e.getMessage());
		} catch (IOException e) {
			WebBrowserApp.debugOut("load Links "+filename+": " + e.getMessage());
		}
		return false;
	}

	static String getFolder(String bookmarkName) {
		if(bookmarkName.startsWith(BOOKMARK_FOLDER_PREFIX)) {
			bookmarkName = bookmarkName.substring(BOOKMARK_FOLDER_PREFIX.length());
			return bookmarkName+BOOKMARK_FOLDER_SEPARATOR;
		}
		int folderEnd = bookmarkName.indexOf(BOOKMARK_FOLDER_SEPARATOR);
		if(folderEnd < 0)
			return null;
		else
			++folderEnd;
		return bookmarkName.substring(0, folderEnd);
	}

	public List<String> getBookmarkFolders() {
		LinkedHashMap<String, URL[]> bookmarks = getBookmarks();
		ArrayList<String> folders = new ArrayList<String>();
		for(Entry<String, URL[]>bookmark : bookmarks.entrySet()) {
			int folderEnd = bookmark.getKey().lastIndexOf(BOOKMARK_FOLDER_SEPARATOR);
			WebBrowserApp.debugOut(bookmark.getKey());
			while(folderEnd > 0) {
				String folder = bookmark.getKey().substring(0, folderEnd+1);
				if(!folders.contains(folder))
					folders.add(folder);
				folderEnd = bookmark.getKey().lastIndexOf(BOOKMARK_FOLDER_SEPARATOR, folderEnd-1);
			}
		}
		for (String folder : new String[] {
				"Apps" + BOOKMARK_FOLDER_SEPARATOR,
				"Apps" + BOOKMARK_FOLDER_SEPARATOR + HMEItem.GENRE_MUSIC + BOOKMARK_FOLDER_SEPARATOR,
				"Apps" + BOOKMARK_FOLDER_SEPARATOR + HMEItem.GENRE_VIDEO + BOOKMARK_FOLDER_SEPARATOR,
				"Apps" + BOOKMARK_FOLDER_SEPARATOR + HMEItem.GENRE_PHOTOS + BOOKMARK_FOLDER_SEPARATOR }) {
			if(!folders.contains(folder))
				folders.add(folder);
		}
		WebBrowserApp.debugOut("Folders: "+folders);
		return folders;
	}

	/**
	 * Convert internal folder or title format to a "folder/folder/title" string
	 * @param folder
	 * @return
	 */
	public static String folderToReadablePath(String folder) {
		if (folder == null)
			return "";
		if (folder.startsWith(BOOKMARK_FOLDER_PREFIX)) {
			folder = folder.substring(BOOKMARK_FOLDER_PREFIX.length());
		}
		return folder.replaceAll(Pattern.quote(BOOKMARK_FOLDER_SEPARATOR), "/");
	}
	
	public static String readablePathToFolder(String path) {
		if (path == null)
			return "";
		return path.replaceAll(Pattern.quote("/"), BOOKMARK_FOLDER_SEPARATOR);
	}
	
	private static void getBookmarks(LinkedHashMap<String, URL[]> result,
			String bookmarkString) {
		String[] bookmarks = bookmarkString.split(BOOKMARK_ENTRY_SEP);
		for (String bookmark : bookmarks) {
			String[] parts = bookmark.split(BOOKMARK_URL_TITLE_SEP);
			if (parts.length == 3 || parts.length == 4) {
				URL url;
				try {
					url = new URL(parts[0]);
				} catch (MalformedURLException e) {
					System.err.println(e.toString());
					continue;
				}
				URL iconurl = null;
				// backward compatibility
				if (parts.length == 4) {
					saveCookie(url, parts[2]);
					if (parts[1] != null && parts[1].trim().length() > 0) {
						iconurl = URLConnectionHandler.getURL(url, parts[1]);
						if(iconurl == null) {
							continue;
						}
					}
				} else
					saveCookie(url, parts[1]);
				result.put(parts[parts.length - 1], new URL[] { url, iconurl });
			}
		}
	}

	private String addBookmark(String bookmarkString, URL urlBase2,
			String titleString2, URL icon) {
		String iconString = getRelativeURL(urlBase2, icon);
		if (iconString == null) {
			iconString = ""; //$NON-NLS-1$
		} else if(iconString.endsWith(".html") || iconString.endsWith(".htm")) {
			//eliminate bogus entries
			iconString = ""; //$NON-NLS-1$
		} else if(iconString.lastIndexOf(".") < iconString.lastIndexOf("/")) { //$NON-NLS-1$ //$NON-NLS-2$
			//eliminate bogus entries (no suffix at end part of URL)
			iconString = ""; //$NON-NLS-1$
		} else
			WebBrowserApp.debugOut("bookmark icon:" + iconString); //$NON-NLS-1$
		String bookmark = urlBase2.toString() + BOOKMARK_URL_TITLE_SEP
				+ iconString + BOOKMARK_URL_TITLE_SEP;
		String cookies = getSavedCookie(urlBase2);
		if (cookies != null) {
			bookmark += cookies;
		}
		bookmark += BOOKMARK_URL_TITLE_SEP + titleString2;
		if (bookmarkString.length() > 0) {
			return bookmarkString + BOOKMARK_ENTRY_SEP + bookmark;
		} else {
			return bookmark;
		}
	}

	public static String getRelativeURL(URL urlBase, URL icon) {
		if (icon == null)
			return null;
		if (urlBase.getProtocol().equals(icon.getProtocol())
				&& urlBase.getHost().equals(icon.getHost())
				&& urlBase.getPort() == icon.getPort()) {
			// TODO should be able to just check urlBase.getAuthority() to
			// include everything at once including userInfo, but I'm not
			// certain that's always set...?
			String base = urlBase.getFile();
			String rel = icon.getFile();
			String newRel = rel;
			int bi = base.indexOf("/"); //$NON-NLS-1$
			int ri = rel.indexOf("/"); //$NON-NLS-1$
			while (bi != -1 && bi == ri
					&& base.substring(0, bi).equals(rel.substring(0, ri))) {
				newRel = rel.substring(ri);
				bi = base.indexOf("/", bi + 1); //$NON-NLS-1$
				ri = rel.indexOf("/", ri + 1); //$NON-NLS-1$
			}
//			debugOut(newRel + " = " + icon + " - " + urlBase); //$NON-NLS-1$ //$NON-NLS-2$
			return newRel;
		}
		return icon.toString();
	}

	private String removeBookmark(String bookmarkString, URL urlBase,
			String titleString) {
		String result = null;
		if (bookmarkString != null) {
			String[] bookmarks = bookmarkString.split(BOOKMARK_ENTRY_SEP);
			for (String bookmark : bookmarks) {
				if (bookmark.startsWith(urlBase.toString()
						+ BOOKMARK_URL_TITLE_SEP)) {
					// skip it.
				} else {
					if (result != null && result.trim().length() != 0) {
						result = result + BOOKMARK_ENTRY_SEP + bookmark;
					} else {
						result = bookmark;
					}
				}
				// String[] parts = bookmark.split("\t");
				// if(parts.length == 2) {
				// }
			}
		}
		if (result == null) {
			return ""; //$NON-NLS-1$
		}
		return result;
	}
	
	/**
	 * Refresh bookmarks after any automatic updates to the list
	 * @return
	 */
	private boolean updateBookmarks() {
		if (!isBookmarksDisabled()) {
			LinkedHashMap<String, URL[]> bookmarks = getBookmarks();
			boolean changed = applyBookmarkUpdates(bookmarks,
					getLastBookmarkDate(getContext()));
			if(changed) {
				setBookmarks(bookmarks);
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Persist bookmarks as exactly this hashmap (plus cookies)
	 * @param bookmarks
	 */
	private void setBookmarks(LinkedHashMap<String, URL[]> bookmarks) {
		if (!isBookmarksDisabled()) {
			String bookmarkString = ""; //$NON-NLS-1$
//					getContext().getPersistentData(BOOKMARKS_KEY);
			for(Entry<String, URL[]> entry : bookmarks.entrySet()) {
				String titleString = entry.getKey();
				URL urlBase = entry.getValue()[0];
				URL icon = entry.getValue()[1];
				bookmarkString = addBookmark(bookmarkString, urlBase, titleString,
						icon);
			}
			getContext().setPersistentData(BOOKMARKS_KEY, bookmarkString);
		}
	}

	void refreshBookmarks() {
		if(isBookmarksDisabled()) return;
		
		String bookmarkString = getContext().getPersistentData(
				BOOKMARKS_KEY);
		if(bookmarkString == null || bookmarkString.trim().length() == 0) {
			bookmarkString = addDefaultBookmarks("");
		}
		refreshBookmarks(bookmarkString);
	}
	
	private String refreshBookmarks(String bookmarkString) {
		if(isBookmarksDisabled()) return bookmarkString;
		String result = "";
		if (bookmarkString == null) {
			bookmarkString = ""; //$NON-NLS-1$
		}
		
		String[] bookmarks = bookmarkString.split(BOOKMARK_ENTRY_SEP);
		for (String bookmark : bookmarks) {
			String[] parts = bookmark.split(BOOKMARK_URL_TITLE_SEP);
			if (parts.length == 3 || parts.length == 4) {
				URL url;
				try {
					url = new URL(parts[0]);
				} catch (MalformedURLException e) {
					System.err.println(e.toString());
					continue;
				}
				URL iconurl = null;
				// backward compatibility
				if (parts.length == 4) {
					if(parts[1] != null && parts[1].trim().length() > 0) {
						iconurl = URLConnectionHandler.getURL(url, parts[1]);
						if(iconurl == null) {
							continue;
						}
					}
				}
				String title = parts[parts.length - 1];
				
				// gets the most recent cookie when it adds the bookmark.
				result = addBookmark(result, url, title, iconurl);
			}
		}
		
		getContext().setPersistentData(BOOKMARKS_KEY, result);
		return result;
	}	
	
	// gmail seems to keep doing about 1.2M for its cookie, but 20M slowed things to a crawl
	private static final int MAX_COOKIE_LENGTH = 2*1024*1024;

	static {
		// turn on automatic cookie handling
		// have to use a custom CookieStore that loads/stores cookies per
		// TiVo simultaneous... for now one per Thread should do it, though that could end up with multiple per TiVo in theory
//		CookieHandler.setDefault(new CookieManager());
		CookieHandler.setDefault(new CookieManager(new ThreadLocalCookieStore(), null)); // COOKIES
	}
	
	public void initBookmarks() {
		// update bookmarks immediately in case there are automated changes
		updateBookmarks();
		//COOKIES
		// that getBookmarks call also loaded cookies from the bookmarks...let's double-check
		CookieStore cookies = getCookieStore();
		for(java.net.URI uri : cookies.getURIs()) {
			List<HttpCookie> cookieSet = cookies.get(uri);
			WebBrowserApp.debugOut("***cookieURI***"+uri);

			for (int i = 0; i < cookieSet.size(); ++i) {
				HttpCookie cookie = cookieSet.get(i);
				// the toString() includes domain, etc only if version is 1... and we keep convincing the system they're version 0 somehow.
				cookie.setVersion(1); // FIXME this is CHANGING the cookie for debug purposes!
				WebBrowserApp.debugOut("***cookie***"+cookie);
			}
		}
	}
	
	public void destroyBookmarks() {
		// make sure all bookmarks have their latest cookies
		WebBrowserApp.debugOut("**** refreshing all bookmarks");
		refreshBookmarks();

		WebBrowserApp.debugOut("**** removing all cookies");
		getCookieStore().removeAll(); //COOKIES
	}
	
	static CookieStore getCookieStore() {
		return ((CookieManager) CookieHandler.getDefault())
				.getCookieStore();
	}

	private static void saveCookie(URL urlBase, String cookie2) {
		if (cookie2 == null || urlBase == null || cookie2.length() == 0)
			return;
		// known degenerative case... giant cookies... 20k seems somewhat reasonable, but 20M isn't
		// that causes major slowdown that breaks other stuff.
		if(cookie2.length() > MAX_COOKIE_LENGTH) {
			WebBrowserApp.debugOut("saveCookie: cookie arg too large to save for "+urlBase+"! "+cookie2.length());
			return;
		}

		java.net.URI uri;
		try {
			uri = urlBase.toURI();
		} catch (URISyntaxException e) {
			WebBrowserApp.debugOut("CAN'T SAVE COOKIE FOR " + urlBase + ": " + e); //$NON-NLS-1$ //$NON-NLS-2$
			return;
		}
		CookieStore cookies = getCookieStore();
//debugOut("*** saveCookie for "+uri+"\n"+cookie2);
		try {
			List<HttpCookie> cookieSet = HttpCookie.parse(cookie2);
			for (HttpCookie cookie : cookieSet) {
				//COOKIES
				
	//			//force cookies to version 1 so we keep path/domain information!
	//			cookie.setVersion(1);
				
				// now that it parsed as version 1 to get multi-cookie, set to version 0 so it works properly on connections(?)
				cookie.setVersion(0);
				cookies.add(uri, cookie);
			}
		} catch(Exception e) {
			WebBrowserApp.debugOut("Error with cookie \""+cookie2+"\" for "+urlBase+":\n"+e);
			return;
		}
	}

	public boolean moveBookmark(String folder, int index, int newIndex, String newTitle) {
		String bookmarkString = getBookmarkData(getContext());
		// normalize the folder
		if (folder == null)
			folder = "";
		java.util.LinkedHashMap<String, URL[]> bookmarks = getBookmarks(folder, bookmarkString, null);
		if(folder != null && folder.startsWith(BOOKMARK_FOLDER_PREFIX)) {
			folder = getFolder(folder);
		}
		// find the bookmark at "index"
		int targeti = 0;
		String targettitle = null;
		URL targeturl = null;
		URL targeticon = null;
		for (Entry<String, URL[]> entry : bookmarks.entrySet()) {
			if (targeti == index) {
				targettitle = folder + entry.getKey();
				targeturl = entry.getValue()[0];
				targeticon = null;
				if (entry.getValue().length > 1)
					targeticon = entry.getValue()[1];
				break;
			}
			++targeti;
		}
		// if we didn't find it, we can't move it.
		if(targettitle == null) {
			return false;
		}
		
		// pre-remove the one we're moving so we don't accidentally remove it from its new location
		bookmarkString = removeBookmark(bookmarkString, targeturl, targettitle);
		
		// make sure new title and folder are set up correctly before we use it.
		if (newTitle == null) {
			newTitle = targettitle;
		} else {
			if (newTitle.indexOf(BOOKMARK_FOLDER_SEPARATOR) < 0) {
				// need to include original folder in newTitle
				newTitle = folder + newTitle;
			}
		}
		
		// move all of this folder to the end of the bookmarks one at a time,
		// moving our target while we go through them.
		int i = 0;
		for(Entry<String, URL[]> entry : bookmarks.entrySet()) {
			String title = entry.getKey();
			if (!title.startsWith(BOOKMARK_FOLDER_PREFIX)) {
				title = folder + title;
			}
			URL url = entry.getValue()[0];
			URL icon = null;
			if(entry.getValue().length > 1)
				icon = entry.getValue()[1];
			// remove the old one. (target item already removed, don't remove it from new location)
			if(i != index)
				bookmarkString = removeBookmark(bookmarkString, url, title);
			// place the target at its new location before the current entry at that location.
			if(i == newIndex) {
				WebBrowserApp.debugOut("place target: "+newTitle);
				bookmarkString = addBookmark(bookmarkString, targeturl, newTitle, targeticon);
			}
			// add each old one at the end except the moved one.
			if (i != index
					&& !(url.toString().startsWith(BOOKMARK_FOLDER_PREFIX)
							|| title.endsWith(BOOKMARK_FOLDER_SEPARATOR))) {
				WebBrowserApp.debugOut("place: "+title);
				bookmarkString = addBookmark(bookmarkString, url, title, icon);
			}
			++i;
		}
		// new index was too high, just add entry at end
		if(i <= newIndex) {
			WebBrowserApp.debugOut("place target at end: "+newTitle);
			bookmarkString = addBookmark(bookmarkString, targeturl, newTitle, targeticon);
		}
		
		//save the bookmarks
		getContext().setPersistentData(BOOKMARKS_KEY, bookmarkString);
	
		return true;
	}

	static String getSavedCookie(URL urlBase) {
		if (urlBase == null)
			return null;

		java.net.URI uri;
		try {
			uri = urlBase.toURI();
		} catch (URISyntaxException e) {
			WebBrowserApp.debugOut("CAN'T get SAVED COOKIE FOR " + urlBase + ": " + e); //$NON-NLS-1$ //$NON-NLS-2$
			return null;
		}
		CookieStore cookies = getCookieStore();
		List<HttpCookie> cookieSet = cookies.get(uri);

		StringBuffer cookieString = new StringBuffer();
		for (int i = 0; i < cookieSet.size(); ++i) {
			if (i != 0) {
				cookieString.append(","); //$NON-NLS-1$
			}
			//COOKIES
			HttpCookie cookie = cookieSet.get(i);
			// the toString() includes domain, etc only if version is 1... and we keep convincing the system they're version 0 somehow.
//			cookie.setVersion(1);
			cookieString.append(cookie);
			// actually, setting it to version 1 doesn't help either...that uses $Path and $Domain which don't get parsed!
			// duplicating code from cookie's toString, but without the $'s
	        if (cookie.getPath() != null)
	        	cookieString.append(";Path=\"").append(cookie.getPath()).append('"');
	        if (cookie.getDomain() != null)
	        	cookieString.append(";Domain=\"").append(cookie.getDomain()).append('"');
	        if (cookie.getPortlist() != null)
				cookieString.append(";Port=\"").append(cookie.getPortlist()).append('"');
	        // we also need to remember some other things for storage
	        if (cookie.getMaxAge() != -1)
	        	cookieString.append(";Max-Age=\"").append(cookie.getMaxAge()).append('"');
	        if (cookie.getSecure())
	        	cookieString.append(";Secure");
		}
		// toString doesn't even include anything that parseCookie needs to figure out this is version 1
		if(cookieSet.size() > 0) {
			cookieString.append(";version=1");
		}
		
		if(cookieString.length() > MAX_COOKIE_LENGTH) {
			WebBrowserApp.debugOut("getSavedCookie: cookie too large to return for "+urlBase+"! "+cookieString.length());
			return "";
		}
//debugOut("*** getSavedCookie for "+uri+"\n"+cookieString);
		
		return cookieString.toString();
	}

	TeDict getSavedCookiesAsTeDict(URL urlBase) {
		if (urlBase == null)
			return null;

		java.net.URI uri;
		try {
			uri = urlBase.toURI();
		} catch (URISyntaxException e) {
			WebBrowserApp.debugOut("CAN'T get SAVED COOKIE FOR " + urlBase + ": " + e); //$NON-NLS-1$ //$NON-NLS-2$
			return null;
		}
		CookieStore cookies = getCookieStore();
		List<HttpCookie> cookieSet = cookies.get(uri);

		TeDict result = new TeDict();
		for (HttpCookie cookie : cookieSet) {
			//FIXME getValue probably isn't the entire cookie spec..
			//   might have to use toString and trim off the "name="
			String cookieString = cookie.toString();
			//FIXME simulator is doc'd to not support quoted strings which this gives you.
			String cookieVal = cookieString.substring(cookieString.indexOf("=")+1); //$NON-NLS-1$
			//cookie.toString());
//			cookie.getValue());
			
			result.add(cookie.getName(), cookieVal);
		}
		return result;
	}

	public void clearCookies() {
		WebBrowserApp.debugOut("**** clearing all cookies");
		getCookieStore().removeAll(); //COOKIES
		refreshBookmarks();
	}	
}
