package com.blackledge.david.tivo;

import java.util.List;
import java.util.Map;

import com.tivo.hme.bananas.BEvent;
import com.tivo.hme.bananas.BView;
import com.tivo.hme.interfaces.ILogger2;
import com.tivo.hme.sdk.Application;
import com.tivo.hme.sdk.HmeEvent;
import com.tivo.hme.sdk.HmeObject;
import com.tivo.hme.sdk.StreamResource;
import com.tivo.hme.sdk.HmeEvent.ResourceInfo;

/**
 * A BView to contain an Application-as-Stream instance (NOTE:
 * Application-as-Stream ONLY WORKS on a BUG-FIXED SDK (parseQuery fixed on the
 * Application class e.g. using {@link V49BApplication}). Once the application
 * starts, it returns control to THIS application, but any events (customizable)
 * focused on this view will be forwarded to the application. Use
 * {@link #startApp(String, Map)} to specify and begin the application, and
 * check {@link #isStarted} to know if the application began OK, or listen for
 * start and end actions you set in {@link #setAppActions(Object, Object)}. You
 * can use {@link #setKeyPressOverrides(List, HmeObject)} and
 * {@link #setConsumingEvents(boolean)} to customize the behavior, or just
 * subclass for more control.
 * 
 * @author David
 * 
 */
public class AppPassThroughView extends BView {

	protected StreamResource appResource;
	private HmeObject keyPressHandler;
	private List<Integer> overrideCodes;
	protected boolean isStarted = false;
	private boolean isConsumingEvents = true;
	private Object startedAction;
	private Object endedAction;
	
	public AppPassThroughView(BView parent, int x, int y, int width,
			int height) {
		super(parent, x, y, width, height);
	}

	/**
	 * Specify the {@link BEvent.Action} values to post when the application
	 * begins and ends (consistent with {@link #isStarted()}). null will mean no
	 * action is posted.
	 * 
	 * @param startedAction
	 * @param endedAction
	 */
	public void setAppActions(Object startedAction, Object endedAction) {
		this.startedAction = startedAction;
		this.endedAction = endedAction;
	}

	/**
	 * Arguments to the method are directly passed to
	 * {@link Application#createStream(String, Map)} (actual instance must be
	 * bug-fixed in parseQuery method!), but a handler is added to notify its
	 * state and return control to this application. This View forward events to
	 * the active application.
	 * 
	 * @param streamedApp
	 * @param map
	 */
	public void startApp(String streamedApp, Map map) {
		if(streamedApp == null) throw new IllegalArgumentException("NULL is not allowed for startApp String argument.");
		
		Application app = getApp();
		// note, we're assuming the application instance's parseQuery
		// implementation is bugfixed... any HMEObject can do the createStream
		// but they all have the parsequery bug.
		try {
			this.appResource = app.createStream(streamedApp, map);
		} catch(NullPointerException npe) {
			throw new IllegalArgumentException(
					"Application Class likely not patched for SDK bug to allow Application As Stream",
					npe);
		}

		this.setResource(this.appResource);
	}
	
	@Override
	public boolean handleEvent(HmeEvent event) {
		switch (event.getOpCode()) {
		case EVT_KEY: {
			if (this.getResource() != null) {
				HmeEvent.Key ir = (HmeEvent.Key) event;
				if (!isOverriddenKey(ir.getCode(), ir.getRawCode())) {
debug("Forward Event "+event);
					this.getResource().sendEvent(event);
					return isConsumingEvents;
				}
debug("Overridden Event "+event);
			}
			break;
		}
		//TODO forward EVT_IDLE, too? probably not EVT_RESOURCE, though.
		// I believe all other events are communicated directly from TiVo to the App, not passing through this parent app.
		case EVT_RSRC_INFO: {
//			debug("* * * RESOURCE INFO " + event); // DEBUG
			break;
		}
		// handle resource status and translate to application start/end actions
		case StreamResource.EVT_RSRC_STATUS: {
//			debug("* * * RESOURCE " + event); // DEBUG

			HmeEvent.ResourceInfo info = (ResourceInfo) event;
			if(info.getMap() != null && info.getMap().get("error.code") != null) {
//					                  int errorCode = Integer.parseInt((String)info.getMap().get("error.code"));
                String errorCode = String.valueOf(info.getMap().get("error.code"));
				String errorText = (String) info.getMap().get(
						"error.text");
				// log it!!!
				getApp().getLogger2().log(ILogger2.LOG_WARNING, "AppPassThroughView Stream Error ("+errorCode+"): "+errorText);
			}
			if (info.getStatus() >= ResourceInfo.RSRC_STATUS_CLOSED) {
				boolean wasStarted = this.isStarted;
				this.isStarted = false;
				
				if(wasStarted && endedAction != null) {
					this.postEvent(new BEvent.Action(this, endedAction));
				}
			}
			else if (info.getStatus() >= ResourceInfo.RSRC_STATUS_READY) {
				boolean wasStarted = this.isStarted;
				this.isStarted = true;
				
				// this is the key line: now we get control back in OUR application
				this.appResource.setActive(false);
				
				if(!wasStarted && startedAction != null) {
					this.postEvent(new BEvent.Action(this, startedAction));
				}
			}
			break;
		}
		}
		return super.handleEvent(event);
	}

	private void debug(String message) {
		getApp().getLogger2().log(ILogger2.LOG_DEBUG, message);
	}

	/**
	 * Simply forward the codes to keyPressHandler if non-null, otherwise
	 * calls super implementation. This method is called before the
	 * contained app is started, otherwise only for key events that return
	 * true in {@link #isOverriddenKey(int, long)}.
	 */
	@Override
	public boolean handleKeyPress(int code, long rawcode) {
		if(keyPressHandler != null) return keyPressHandler.handleKeyPress(code, rawcode);
		return super.handleKeyPress(code, rawcode);
	}
	
	/**
	 * Return true if you want to not pass this key through to the
	 * application. Before the application is started, this doesn't apply
	 * (all keys are handled by the usual view handlers)
	 * 
	 * @param code
	 * @param rawCode
	 * @return false if overrideCodes is null or overrideCodes doesn't
	 *         contain the code.
	 */
	protected boolean isOverriddenKey(int code, long rawCode) {
		if(this.overrideCodes != null && this.overrideCodes.contains(code)) return true;
		return false;
	}
	
	/**
	 * 
	 * @param codes if non-null, any KEY event not in this list is forwarded to the app if it is started
	 * @param keyPressHandler if non-null, handles this view's keyPress events only.
	 */
	public void setKeyPressOverrides(List<Integer> codes, HmeObject keyPressHandler) {
		this.keyPressHandler = keyPressHandler;
		this.overrideCodes = codes;
	}

	/**
	 * True by default - if true events forwarded to the contained application
	 * do are considered "handled" and don't bubble up the view hierarchy. If
	 * you want to do additional handling in addition to what the contained app
	 * does, set this to false.
	 * 
	 * @param consume
	 */
	public void setConsumingEvents(boolean consume) {
		this.isConsumingEvents = consume;
	}
	
	public boolean isConsumingEvents() {
		return this.isConsumingEvents;
	}
	
	/**
	 * Has the app finished loading?
	 * @return
	 */
	public boolean isStarted() {
		return this.isStarted;
	}
}