package com.blackledge.david.tivo;

import com.tivo.hme.interfaces.IHmeConstants;
import com.tivo.hme.sdk.IHmeProtocol;

/**
 * IHmeProtocol extension that supports some features beyond SDK 1.4.1
 * especially HTTP Header strings and Slide Remote key codes. Typically just use
 * {@link #KEY_OPT_ASCII_START}, {@link #KEY_OPT_ASCII_END}, and
 * {@link #KEY_OPT_ASCII_OFFSET} to get all the characters.
 * 
 */
public interface V49IHmeProtocol extends IHmeProtocol {
	/**
	 * 0.49 - Slide Remote key events supported (no "supporting" sdk version)
	 * (0.41 ({@link IHmeProtocol#VERSION_0_41}) supported by sdk 1.4, 0.43-0.45
	 * ({@link IHmeProtocol#VERSION_0_43}-{@link IHmeProtocol#VERSION_0_45}) by
	 * sdk 1.4.1e)
	 */
	public static final int VERSION_0_49 = (0 << 8) | 49;

	/**
	 * according to an error logged:
	 * "rendered text exceeded max allowed dimensions of 1920x1080"
	 */
	public static final int LIMIT_TEXT_RENDER_WIDTH = 1920;
	/**
	 * according to an error logged:
	 * "rendered text exceeded max allowed dimensions of 1920x1080"
	 */
	public static final int LIMIT_TEXT_RENDER_HEIGHT = 1080;
	
	/** Media status code, resource at end of stream. */
	public static final int RSRC_STATUS_END = 11;
	/** Media status code, resource at end of available buffer. */
	public static final int RSRC_STATUS_BUFFER_OVERFLOW = 12;
	
    /**
	 * Zoom key, optional, relabel of {@link #KEY_OPT_ASPECT} key, which is same
	 * code as {@link #KEY_OPT_WINDOW} and {@link #KEY_OPT_PIP} (22).
	 */
    public static final int KEY_OPT_ZOOM			= KEY_OPT_WINDOW;
    
    // undefined keys in SDK: 0-21, 23, 25, 28-50
    
    /**
	 * On Demand key, reserved for internal use. E.g. long blue ON DEMAND button
	 * on Charter remote. This is sent in a key press event just before
	 * attempting to "teleport" to the service. In anything below Series 4, no
	 * teleport and not transmitted over HME, apparently.
	 */
    public static final int KEY_OPT_ON_DEMAND = 30;
    
    // last SDK key code defined: KEY_OPT_DVD = 55;
    
	/**
	 * "A" yellow action remote button, optional (first on Premiere), may only transmit on if identified as V49.
	 */
	public static final int KEY_OPT_ACTION_A = 56;
	/**
	 * "B" blue action remote button, optional (first on Premiere), may only transmit on if identified as V49.
	 */
	public static final int KEY_OPT_ACTION_B = 57;
	/**
	 * "C" red action remote button, optional (first on Premiere), may only transmit on if identified as V49.
	 */
	public static final int KEY_OPT_ACTION_C = 58;
	/**
	 * "D" green action remote button, optional (first on Premiere), may only transmit on if identified as V49.
	 */
	public static final int KEY_OPT_ACTION_D = 59;
	
	/**
	 * Backspace (picture of a left arrow on slide remote) button code.<BR>
	 * Name matches known network interface KEYBOARD command.
	 */
	public static final int KEY_OPT_BACKSPACE = 65;
	
    /**
     * (TBD (backspace?) requires a higher version than 49) Back key, optional (first on Roamio remote).  Other "Back" keys in most apps: {@link #KEY_OPT_ACTION_B}, {@link #KEY_OPT_ZOOM}, and sometimes KEY_REPLAY.
     */
    public static final int KEY_OPT_BACK = KEY_OPT_BACKSPACE;
    
	// codes 60-64 and 66-69 are unknown, currently
	
	/**
	 * Code returned by the RETURN key on the keyboard if it is modified by Alt
	 * (or Windows or Ctrl or Shift (or possibly "sym"(bol))). Normally that key
	 * sends KEY_SELECT.
	 */
	public static final int KEY_OPT_META_SELECT = 70;
	/**
	 * Left Shift button code. You do not need to process this to handle
	 * capitalization, use KEY_OPT_ASCII_* to get the properly capitalized
	 * character.
	 * Name matches known network interface KEYBOARD command
	 */
	public static final int KEY_OPT_LSHIFT = 71;
	/**
	 * Right Shift button code. You do not need to process this to handle
	 * capitalization, use KEY_OPT_ASCII_* to get the properly capitalized
	 * character.
	 * Name matches known network interface KEYBOARD command
	 */
	public static final int KEY_OPT_RSHIFT = 72;
	/**
	 * Left Alt (and Windows) key button code for keyboards - POSSIBLY the
	 * "sym"(bol) key for the Slide Remote (although there is only one, on the
	 * right side). You do not need to process this to handle symbols, use
	 * KEY_OPT_ASCII_* to get the properly capitalized character.
	 * Name matches known network interface KEYBOARD command
	 */
	public static final int KEY_OPT_LMETA = 73;
	/**
	 * Right Alt (and Windows) key button code for keyboards - POSSIBLY the
	 * "sym"(bol) key for the Slide Remote. You do not need to process this to
	 * handle symbols, use KEY_OPT_ASCII_* to get the properly capitalized
	 * character.
	 * Name matches known network interface KEYBOARD command
	 */
	public static final int KEY_OPT_RMETA = 74;
	/**
	 * Capslock button code. You do not need to process this to handle
	 * capitalization, use KEY_OPT_ASCII_* to get the properly capitalized
	 * character.
	 * Name matches known network interface KEYBOARD command
	 */
	public static final int KEY_OPT_CAPS = 75;

	/** known from network interface KEYBOARD command, just sends non-KBD code on a USB keyboard */
	public static final int KEY_OPT_KBDUP = KEY_UP;
	/** known from network interface KEYBOARD command, just sends non-KBD code on a USB keyboard */
	public static final int KEY_OPT_KBDDOWN = KEY_DOWN;
	/** known from network interface KEYBOARD command, just sends non-KBD code on a USB keyboard */
	public static final int KEY_OPT_KBDLEFT = KEY_LEFT;
	/** known from network interface KEYBOARD command, just sends non-KBD code on a USB keyboard */
	public static final int KEY_OPT_KBDRIGHT = KEY_RIGHT;
	/** known from network interface KEYBOARD command, just sends SELECT on a USB keyboard */
	public static final int KEY_OPT_KBDENTER = KEY_SELECT;
	/** known from network interface KEYBOARD command, just sends CLEAR on a USB keyboard */
	public static final int KEY_OPT_ESCAPE = KEY_CLEAR;
	
	/** NO keycode sent in V49 HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_LCONTROL = KEY_UNKNOWN;
	/** NO keycode sent in V49 HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_RCONTROL = KEY_UNKNOWN;
	/** NO keycode sent in V49 HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_PAGEUP = KEY_UNKNOWN;
	/** NO keycode sent in V49 HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_PAGEDOWN = KEY_UNKNOWN;
	/** NO keycode sent in V49 HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_HOME = KEY_UNKNOWN;
	/** NO keycode sent in V49 HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_INSERT = KEY_UNKNOWN;
	/** NO keycode sent in V49 HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_DELETE = KEY_UNKNOWN;
	
	/** UNKNOWN keycode (same as ADVANCE?) - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_DELIMITER = KEY_ADVANCE;
	
	/** UNKNOWN keycode - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_CC_ON = KEY_UNKNOWN;
	/** UNKNOWN keycode - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_CC_OFF = KEY_UNKNOWN;
	/** UNKNOWN keycode - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_ASPECT_CORRECTION_FULL = KEY_UNKNOWN;
	/** UNKNOWN keycode - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_ASPECT_CORRECTION_PANEL = KEY_UNKNOWN;
	/** UNKNOWN keycode - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_ASPECT_CORRECTION_ZOOM = KEY_UNKNOWN;
	/** UNKNOWN keycode - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_ASPECT_CORRECTION_WIDE_ZOOM = KEY_UNKNOWN;
	/** UNKNOWN keycode - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_VIDEO_MODE_FIXED_480i = KEY_UNKNOWN;
	/** (teleport key) UNKNOWN keycode - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_STANDBY = KEY_UNKNOWN;
	/** (teleport key) reported to match TIVO key... unless it matches {@link #KEY_OPT_ON_DEMAND} - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_DIRECTV = KEY_TIVO;
	/** (teleport key) UNKNOWN keycode - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_NOWSHOWING = KEY_UNKNOWN;
	/** (teleport key) UNKNOWN keycode (same as NOWSHOWING?) - may not be sent through HME, known from network interface KEYBOARD command */
	public static final int KEY_OPT_LIST = KEY_OPT_NOWSHOWING;
	
	/**
	 * For all characters between KEY_OPT_ASCII_START and KEY_OPT_ASCII_END
	 * (inclusive) subtract this value from the key code to get the ASCII
	 * character they represent.
	 */
	public static final int KEY_OPT_ASCII_OFFSET = 0x10000; // 65536, 2^16
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_SPACE = KEY_OPT_ASCII_OFFSET+' ';
	// Sym Q
	public static final int KEY_OPT_ASCII_EXCLAMATION = KEY_OPT_ASCII_OFFSET+'!';
	// Sym L
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_DOUBLEQUOTE = KEY_OPT_ASCII_OFFSET+'\"';
	// Sym E
	public static final int KEY_OPT_ASCII_HASH = KEY_OPT_ASCII_OFFSET+'#';
	// Sym R
	public static final int KEY_OPT_ASCII_DOLLAR = KEY_OPT_ASCII_OFFSET+'$';
	// Sym T
	public static final int KEY_OPT_ASCII_PERCENT = KEY_OPT_ASCII_OFFSET+'%';
	// Sym U
	public static final int KEY_OPT_ASCII_AMPERSAND = KEY_OPT_ASCII_OFFSET+'&';
	// Sym K
	/** (Apostrophe) name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_QUOTE = KEY_OPT_ASCII_OFFSET+'\'';
	// Sym O
	public static final int KEY_OPT_ASCII_LPAREN = KEY_OPT_ASCII_OFFSET+'(';
	// Sym P
	public static final int KEY_OPT_ASCII_RPAREN = KEY_OPT_ASCII_OFFSET+')';
	// Sym I
	public static final int KEY_OPT_ASCII_ASTERISK = KEY_OPT_ASCII_OFFSET+'*';
	// Sym 1
	public static final int KEY_OPT_ASCII_PLUS = KEY_OPT_ASCII_OFFSET+'+';
	// Sym B
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_COMMA = KEY_OPT_ASCII_OFFSET+',';
	// Sym 2
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_MINUS = KEY_OPT_ASCII_OFFSET+'-';
	// Sym N
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_PERIOD = KEY_OPT_ASCII_OFFSET+'.';
	// Sym 5
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_SLASH = KEY_OPT_ASCII_OFFSET+'/';
	// 0-9 use KEY_NUM0-KEY_NUM9
	// Sym J
	public static final int KEY_OPT_ASCII_COLON = KEY_OPT_ASCII_OFFSET+':';
	// Sym H
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_SEMICOLON = KEY_OPT_ASCII_OFFSET+';';
	// Sym Z
	public static final int KEY_OPT_ASCII_LESS = KEY_OPT_ASCII_OFFSET+'<';
	// Sym 3
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_EQUALS = KEY_OPT_ASCII_OFFSET+'=';
	// Sym X
	public static final int KEY_OPT_ASCII_GREATER = KEY_OPT_ASCII_OFFSET+'>';
	// Sym M
	public static final int KEY_OPT_ASCII_QUESTION = KEY_OPT_ASCII_OFFSET+'?';
	// Sym W
	public static final int KEY_OPT_ASCII_AT = KEY_OPT_ASCII_OFFSET+'@';
	public static final int KEY_OPT_ASCII_A_UPPER = KEY_OPT_ASCII_OFFSET+'A';
	public static final int KEY_OPT_ASCII_Z_UPPER = KEY_OPT_ASCII_OFFSET+'Z';
	// Sym F
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_LBRACKET = KEY_OPT_ASCII_OFFSET+'[';
	// Sym V
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_BACKSLASH = KEY_OPT_ASCII_OFFSET+'\\';
	// Sym G
	/** name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_RBRACKET = KEY_OPT_ASCII_OFFSET+']';
	// Sym Y
	public static final int KEY_OPT_ASCII_CIRCUMFLEX = KEY_OPT_ASCII_OFFSET+'^';
	// Sym 7
	public static final int KEY_OPT_ASCII_UNDERSCORE = KEY_OPT_ASCII_OFFSET+'_';
	// Sym 9
	/** (Grave accent mark) name matches known network interface KEYBOARD command */
	public static final int KEY_OPT_ASCII_BACKQUOTE = KEY_OPT_ASCII_OFFSET+'`';
	public static final int KEY_OPT_ASCII_A_LOWER = KEY_OPT_ASCII_OFFSET+'a';
	public static final int KEY_OPT_ASCII_Z_LOWER = KEY_OPT_ASCII_OFFSET+'z';
	// Sym S
	public static final int KEY_OPT_ASCII_LBRACE = KEY_OPT_ASCII_OFFSET+'{';
	// Sym C
	public static final int KEY_OPT_ASCII_BAR = KEY_OPT_ASCII_OFFSET+'|';
	// Sym D
	public static final int KEY_OPT_ASCII_RBRACE = KEY_OPT_ASCII_OFFSET+'}';
	// Sym A
	public static final int KEY_OPT_ASCII_TILDE = KEY_OPT_ASCII_OFFSET+'~';
	/**
	 * space character - beginning of range of values that are ASCII characters
	 * plus KEY_OPT_ASCII_OFFSET. Note that 0-9 do not show up this way,
	 * instead sending the KEY_NUM* codes.
	 */
	public static final int KEY_OPT_ASCII_START = KEY_OPT_ASCII_SPACE;
	/**
	 * tilde character - end of range of values that are ASCII characters plus
	 * KEY_OPT_ASCII_OFFSET. Note that 0-9 do not show up this way,
	 * instead sending the KEY_NUM* codes.
	 */
	public static final int KEY_OPT_ASCII_END = KEY_OPT_ASCII_TILDE;

	// as discovered by davidblackledge - Premiere only.
	public static final int ID_FONT3_TTF = 13;
	public static final int ID_FONT3ITALIC_TTF = 14;
	public static final int ID_FONT3BOLD_TTF = 15;
	// as discovered by wmcbrine - Premiere only
	public static final int ID_SPEEDUP4_SOUND = 37;
	
	/**
	 * HTTP Header sent in resource requests by the TiVo with the tsn (context
	 * connection attribute "tsn") as value.
	 */
	public static final String HTTP_HEADER_TIVO_ID = "TiVo_TCD_ID";

	/**
	 * HTTP Header sent in resource requests by the TiVo with the TiVo Software
	 * Version (Device Info event's "version") as value.
	 */
	public static final String HTTP_HEADER_TIVO_VERSION = "TiVo_SW_VER";

	public static final String HTTP_HEADER_TIVO_DURATION = IHmeConstants.TIVO_DURATION;

}
