package com.blackledge.david.tivo.game;

import java.awt.Color;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import com.blackledge.david.game.Chat;
import com.blackledge.david.game.GameException;
import com.blackledge.david.game.MultiplayerGame;
import com.blackledge.david.game.MultiplayerGame.GameData;
import com.blackledge.david.game.PlayerState;
import com.blackledge.david.game.MultiplayerGame.GameInfo;

import com.blackledge.david.tivo.RemotePicker;
import com.blackledge.david.tivo.TiVoModelInfo;
import com.blackledge.david.tivo.V49BApplication;
import com.blackledge.david.tivo.V49BApplicationFactory;

import com.tivo.hme.bananas.BView;
import com.tivo.hme.sdk.HmeEvent;
import com.tivo.hme.sdk.Resource;
import com.tivo.hme.sdk.StreamResource;

public abstract class Game extends V49BApplication {
	public static abstract class GameFactory extends V49BApplicationFactory 
	{

		public int getBoardHeight() {
			return getGame().getBoardHeight();
		}

		public int getBoardWidth() {
			return getGame().getBoardWidth();
		}

		/** Instantiate game data, and return id for that instantiation */
		public synchronized int newGame() {
			return getGame().newGame();
		}
		
		public abstract MultiplayerGame getGame();

		/**
		 * Generic representation of a player doing something. It may or may not
		 * be successful depending on the rules of the game. A successful action
		 * may or may not affect the game state.
		 * 
		 * @param gameID
		 * @param playerIdentification
		 * @param action
		 * @param details
		 * @return true if the action was successful, false if it was not
		 *         allowed.
		 */
		public boolean playerAction(int gameID,
				PlayerState playerIdentification, int action, Object details) {
			return getGame().playerAction(gameID, playerIdentification, action, details);
		}
		
//		/**
//		 * Initialize a new game before players are added.  Mainly set game.boardState to a new instance
//		 * @param game
//		 */
//		protected abstract void initNewGame(GameData game);

		/**
		 * Add a player to a game and update state (e.g. start game) as appropriate.
		 * Throw full or id_in_use exceptions, modify player state or game as
		 * needed, and add player to game.players. Default impl throws
		 * ID_IN_USE if player is already in players, GAME_FULL if players already has maximumPlayers, otherwise adds player to
		 * game.players and if this player makes game meet minimumPlayers, FIRST player gets turn.
		 * Also chats "ENTERED"
		 * @param game
		 * @param player
		 * @throws GameException
		 */
		protected void addPlayer(GameData game, PlayerState player) throws GameException {
			// check the game isn't full of players already
			if(game.players.size() >= game.maximumPlayers) {
				throw GameException.GAME_FULL;
			}
			
			// check the playerid isn't already in the game
			if(game.players.contains(player))
				throw GameException.ID_IN_USE;

			game.players.add(player);
			
			playerChat(game.gameID, player, "ENTERED");
			
			if(game.players.size() == game.minimumPlayers) {
				//TODO if player 0 already HAS turn.. (leftover from abandonment) should somebody else get it to make sure turns stay balanced?
				game.players.get(0).hasTurn = true;
			}
		}

		/**
		 * Remove a player from a game and update state (e.g. assign remaining
		 * player as winner) as appropriate. remove player from game.players.
		 * Default Impl just removes player from game.players, first runs {@link #nextTurn(GameData, PlayerState)} if this player hasTurn.
		 * Also chat's "EXITED"
		 * @param game
		 * @param player
		 */
		protected void removePlayer(GameData game, PlayerState player) {
			// if player with turn is leaving, give turn to somebody remaining
			if(player.hasTurn) {
				nextTurn(game, player);
			}
			game.players.remove(player);
			
			playerChat(game.gameID, player, "EXITED");
		}

		/**
		 * change entries to move turn from player to next entry in game.players.
		 * @param game
		 * @param player
		 */
		protected void nextTurn(GameData game, PlayerState player) {
			int p1index = game.players.indexOf(player);
			int p2index;
			if(game.players.size() > p1index+1) {
				p2index = p1index+1;
			} else {
				p2index = 0;
			}
			player.hasTurn = false;
			game.players.get(p1index).hasTurn = false;
			
			game.players.get(p2index).hasTurn = true;
		}
		
		/**
		 * Get the FIRST player that has turn.
		 * @param game
		 * @return
		 */
		protected PlayerState getPlayerWithTurn(GameData game) {
			for(PlayerState player : game.players) {
				if(player.hasTurn) {
					return player;
				}
			}
			return null;
		}
		
		protected List<PlayerState> getPlayersWithTurn(GameData game) {
			ArrayList<PlayerState> result = new ArrayList<PlayerState>();
			for(PlayerState player : game.players) {
				if(player.hasTurn) {
					result.add(player);
				}
			}
			return result;
		}
		
		public void playerChat(int gameID, PlayerState speaker, String message) {
			Date timestamp = new Date();
			GameData gameData = getGame().getGameData(gameID);
			gameData.chat.add(new Chat(speaker, timestamp, message));
			gameData.lastChanged = timestamp;
			gameData.lastChangedChat = timestamp;
		}

		public void playerUpdate(int gameID, PlayerState player) {
			Date timestamp = new Date();
			GameData gameData = getGame().getGameData(gameID);
			int idx = gameData.players.indexOf(player);
			gameData.players.set(idx, new PlayerState(player));
			gameData.lastChanged = timestamp;
			gameData.lastChangedPlayers = timestamp;
		}

		private List<PlayerState> _connectedPlayers = new ArrayList<PlayerState>();
		public List<PlayerState> getConnectedPlayers() {
			synchronized (_connectedPlayers) {
				return new ArrayList<PlayerState>(_connectedPlayers);
			}
		}
		public List<PlayerState> getIdlePlayers() {
			List<PlayerState> players = getConnectedPlayers();
			for(GameInfo game : getGame().getGameList()) {
				players.removeAll(game.players);
			}
			return players;
		}

		public void connectPlayer(PlayerState player) {
			synchronized (_connectedPlayers) {
				if (!_connectedPlayers.contains(player)) {
					_connectedPlayers.add(player);
				}
			}
		}

		public void disconnectPlayer(PlayerState player) {
			synchronized (_connectedPlayers) {
				_connectedPlayers.remove(player);
			}
		}

	}
	

	
	
	protected PlayerState _player;
	public PlayerState getPlayer() {
		return _player;
	}
	
	protected int _gameID;

	public int getGameID() {
		return _gameID;
	}
	public void setGameID(int gameID) {
		_gameID = gameID;
	}
	

	/**
	 * Override this and call this implementation first (which sets app._player
	 * based on this.tivoID and connects the user to the game server) then you
	 * must push a custom instance of GameLobby
	 */
	@Override
	protected void showFirstPage() {
		String name = null;
		// prefer "host" as name.
		if (this.deviceInfo != null) {
			Object host = this.deviceInfo.get("host");
			if (host != null & String.valueOf(host).trim().length() > 0) {
				name = String.valueOf(host);
			}
		}
		TiVoModelInfo model = TiVoModelInfo.getInfo(this.tivoID);
		String modelsuffix = this.tivoID.substring(this.tivoID.length() - 4);
		// default to model-based name.
		if (name == null) {
			name = (model == null) ? this.tivoID : model.getSeriesName() + "-"
					+ modelsuffix;
		} else {
			// make sure unique number is part of name.
			if(name.indexOf(modelsuffix) < 0) {
				name = name + " " + modelsuffix;
			}
		}
		_player = new PlayerState(this.tivoID, name);
		GameFactory gameFactory = (GameFactory) getFactory();
		gameFactory.connectPlayer(_player);

		mediaInit();
		
		push(newGameLobby(this), TRANSITION_FORWARD);
	}
	
	@Override
	protected void destroyCleanup() {
		GameFactory gameFactory = (GameFactory) getFactory();
		try {
			gameFactory.getGame().leaveGame(_gameID, _player);
		} catch (GameException e) {
			if(e.equals(GameException.GAME_NOT_FOUND)) {
				// expected
			} else {
			// TODO Auto-generated catch block
			e.printStackTrace();
			}
		}
		gameFactory.disconnectPlayer(_player);

		super.destroyCleanup();
	}

	private boolean _stopLobbyUpdate;
	protected void startLobbyUpdates() {
		updateLobbyNow();
		_stopLobbyUpdate = false;
		scheduleLobbyUpdate(0, 1000);
	}
	protected void stopLobbyUpdates() {
		_stopLobbyUpdate = true;
	}
	protected void updateLobbyNow() {
		processLobbyUpdate(0);
	}

	private boolean _stopPoll;
	protected void startGameUpdates() {
		// get starting game state
		updateGameNow();
		// schedule the first update
		_stopPoll = false;
		schedulePoll(_gameID, 1000);
	}
	protected void stopGameUpdates() {
		_stopPoll = true;
	}

	protected void updateGameNow() {
		processPoll(_gameID);
	}
	

	private Date _lastUpdated;



	private boolean _spectating;

	
	
	private static final int POLL_KEY = KEY_LIVETV;
	private static final int POLL_TIME_DEFAULT = 500;

	private static final int LOBBY_UPDATE_KEY = KEY_TIVO;
	private static final int LOBBY_UPDATE_TIME_DEFAULT = 2000;
	
	@Override
	public boolean handleEvent(HmeEvent event) {
		switch (event.getOpCode()) {
		case EVT_KEY:
			HmeEvent.Key e = (HmeEvent.Key)event;
			if (e.getAction() == KEY_REPEAT) {
				switch (e.getCode()) {
				case POLL_KEY:
				{
					int arg = (int) e.getRawCode();
					int pollTime = processPoll(arg);
					//...repeat
					if(_stopPoll) {
						// when next we start, start from scratch.
						_lastUpdated = null;
//						_stopPoll = false;
						return true;
					}
					if(pollTime <= 0) {
						pollTime = POLL_TIME_DEFAULT;
					}
//					schedulePoll(arg, pollTime);
					schedulePoll(_gameID, pollTime);
					return true;
				}
				case LOBBY_UPDATE_KEY: {
					int arg = (int) e.getRawCode();
					int lobbyUpdateTime = processLobbyUpdate(arg);
					//...repeat
					if(_stopLobbyUpdate) {
//						_stopLobbyUpdate = false;
						return true;
					}
					if(lobbyUpdateTime <= 0) {
						lobbyUpdateTime = LOBBY_UPDATE_TIME_DEFAULT;
					}
					scheduleLobbyUpdate(arg, lobbyUpdateTime);
					return true;
				}
				}
			}
			break;
		}
		return super.handleEvent(event);
	}
	
	@Override
	public boolean handleKeyPress(int code, long rawcode) {
		switch(code) {
		case KEY_SLOW:
			mediaAction();
			return true;
		case KEY_PAUSE:
			mediaPause();
			return true;
		case KEY_OPT_ZOOM:
			mediaZoom();
			return true;
		}
		return super.handleKeyPress(code, rawcode);
	}

	private BView _media;
	private BView _mediaZoomHint;
	private BView _mediaActionHint;
	private boolean _mediaPIP = true;
	private void mediaPause() {
		if(_media != null) {
			if(_media.getResource() != null) {
				if (_media.getResource() instanceof StreamResource) {
					StreamResource r = (StreamResource) _media.getResource();
					r.pause();
				}
			}
		}
	}
	private void mediaAction() {
		if(_media == null) {
			mediaCreatePIP();
		}
		
		if(_media.getVisible()) {
			mediaHide();
		} else {
			mediaPIP();
			mediaShow();
		}
	}
	private void mediaZoom() {
		if(_media.getVisible()) {
			if(_mediaPIP) {
				mediaFullScreen();
			} else {
				mediaPIP();
			}
		}
	}
	
	private void mediaShow() {
		_media.setVisible(true);
		_media.setResource(createStream(V49BApplication.STREAM_LIVETV, "video/mpeg", true));
	}
	
	private void mediaFullScreen() {
		scalePIP(false);

		mediaHintPosition(false);
	}
	private void mediaPIP() {
		scalePIP(true);
		
		mediaHintPosition(true);
	}
	private void mediaHintPosition(boolean pip) {
		float transparency = 0.2f;
		if(!pip) {
			transparency = 0.0f;
		}
		Resource anim = getResource("*100");
		_mediaZoomHint.setLocation(
				_media.getX()
				- _mediaZoomHint.getWidth(), 
				(int) (_media.getY()+(_media.getHeight()*_media.getScaleY()))
				- _mediaZoomHint.getHeight(),
				anim);
		_mediaZoomHint.setTransparency(transparency, anim);
		_mediaZoomHint.setVisible(true, anim);
		_mediaActionHint.setLocation(
				_media.getX()
				- _mediaActionHint.getWidth(), 
				_media.getY()
				+ _mediaActionHint.getHeight()/2, //FIXME too far (overlaps with zoom hint) in PIP + SD
				anim);
		_mediaActionHint.setTransparency(transparency, anim);
		_mediaActionHint.setVisible(true, anim);
	}
	private void mediaHide() {
		_media.setVisible(false);
		if(_media.getResource() != null) {
			_media.getResource().remove();
			_media.clearResource();
		}
		Resource anim = getResource("*100");
		_mediaActionHint.setLocation(getAbove().getWidth()-getSafeTitleHorizontal(), getSafeTitleVertical(),
				anim);
		_mediaActionHint.setTransparency(0.0f, anim);
		_mediaZoomHint.setLocation(getAbove().getWidth()-getSafeTitleHorizontal(), getSafeTitleVertical(),
				anim);
		_mediaZoomHint.setTransparency(0.0f, anim);
		_mediaZoomHint.setVisible(false);
	}
	
	private void mediaInit() {
		// make media area before hints so hints will appear over it
		mediaCreatePIP();
		
		int width = 50;
		Color textColor = Color.green.brighter();
		int textHeight = 20;
		String font = "default-15.font";
		Rectangle region = new Rectangle(getAbove().getWidth()-getSafeTitleHorizontal(), getSafeTitleVertical(), width, (int) (width*1.5));
		_mediaActionHint = new BView(getAbove(), region.x, region.y, region.width, region.height);
		_mediaActionHint.setResource(Color.DARK_GRAY);
		RemotePicker.addHint(RemotePicker.getRemoteForTiVo(this.tivoID), _mediaActionHint, 
//				new int[] {KEY_SLOW}, "TV or \nMusic", font, textHeight, textColor, width, false);
				new int[] {KEY_SLOW}, "Stream\nTV", font, textHeight, textColor, width, false);
		_mediaZoomHint = new BView(getAbove(), region.x, region.y, region.width, region.height, false);
		_mediaZoomHint.setResource(Color.DARK_GRAY);
		RemotePicker.addHint(RemotePicker.getRemoteForTiVo(this.tivoID), _mediaZoomHint, 
				new int[] {KEY_OPT_ZOOM}, "PIP /\nFull", font, textHeight, textColor, width, false);
	}
	
	private void mediaCreatePIP() {
		_media = new BView(getAbove(), 0, 0, getWidth(), getHeight(), false);
		scalePIP(true);
	}

	// copied/adapted from VideHandlerView
	public static final float PIP_SCALE = 2f/7f; // I measured Premiere's pip as > 2/7ths, <.3 by width
	public static final int PIP_EDGE_OFFSET = 10; // PIP must be away from edge or scale is ignored by video.
	private void scalePIP(boolean pip) {
		synchronized(this) {
			boolean wasPainting = _media.getPainting();
			_media.setPainting(false);
			this._mediaPIP = pip;
			if (pip) {
				float pipScale = PIP_SCALE;
				int pipWidth = (int) (getWidth() * pipScale);
				int edgeOffset = PIP_EDGE_OFFSET;
				_media.setLocation(getWidth()-pipWidth-edgeOffset, edgeOffset, null);
				flush(); try { Thread.sleep(100); } catch(Exception e) {}
				_media.setScale(pipScale, pipScale);
			} else {
//				//Full screen
//				_media.setLocation(0, 0);
//				flush(); try { Thread.sleep(100); } catch(Exception e) {}
//				_media.setScale(1.0f, 1.0f);
				//Framed
				_media.setLocation(getSafeTitleHorizontal(), getSafeTitleVertical());
				flush(); try { Thread.sleep(100); } catch(Exception e) {}
				_media.setScale(((float)getWidth() - (float)getSafeTitleHorizontal()*2.0f) / (float)getWidth(),
						((float)getHeight() - (float)getSafeTitleVertical()*2.0f) / (float)getHeight());

			}
			if (wasPainting) {
				_media.setPainting(true);
			}
		}
	}
	

	private void schedulePoll(int arg, int pollTime) {
		sendEvent(new HmeEvent.Key(getApp().getID(), KEY_REPEAT, POLL_KEY, arg), getResource("*"+pollTime));
	}
	
	private int processPoll(int arg) {
		if(getCurrentScreen() != null && getCurrentScreen() instanceof GameScreen && _gameID >= 0) {
			GameScreen screen = (GameScreen) getCurrentScreen();
			GameFactory gameFactory = (GameFactory) getFactory();
			GameData game = gameFactory.getGame().getGameData(_gameID);
			if (game != null && (_lastUpdated == null
					|| gameFactory.getGame().getGameUpdated(_gameID).after(_lastUpdated))) {
	
				if(_lastUpdated == null || gameFactory.getGame().getPlayersUpdated(_gameID).after(_lastUpdated)) {
					screen.updateGamePlayers(game.players);
				}
				if(_lastUpdated == null || gameFactory.getGame().getChatUpdated(_gameID).after(_lastUpdated)) {
					screen.updateGameChat(game.chat);
				}
				if(_lastUpdated == null || gameFactory.getGame().getBoardUpdated(_gameID).after(_lastUpdated)) {
					screen.updateGameBoard(game.boardState);
				}
				
				_lastUpdated = new Date();
			}
//		else
//			_lastUpdated = new Date();
		}
		
		return 0;
	}

	private void scheduleLobbyUpdate(int arg, int pollTime) {
		sendEvent(new HmeEvent.Key(getApp().getID(), KEY_REPEAT, LOBBY_UPDATE_KEY, arg), getResource("*"+pollTime));
	}
	
	private int processLobbyUpdate(int arg) {
		if(getCurrentScreen() != null && getCurrentScreen() instanceof GameLobby) {
			GameLobby screen = (GameLobby) getCurrentScreen();
			screen.update();
		}
		
		return 0;
	}
	
	/**
	 * Instantiate the GameLobby instance appropriate to your game
	 */
	protected abstract GameLobby newGameLobby(Game app);

	/**
	 * instantiates and pushes the GameScreen, representing the joining of this
	 * user into the identified game.
	 * 
	 * @param gameID
	 *            the id of the game... negative values are flags for special
	 *            cases.
	 * @param spectate
	 *            true if the user should be joined as a spectator, not a player
	 */
	public abstract void joinGame(int gameID, boolean spectate);
	
	public void setSpectating(boolean spectate) {
		_spectating = spectate;	
	}
	public boolean isSpectating() {
		return _spectating;
	}



}
